@@ -1,4 +1,4 @@
-/* $EPIC: irc.c,v 1.777 2014/02/11 20:42:38 jnelson Exp $ */
+/* $EPIC: irc.c,v 1.1336 2012/08/05 01:28:03 jnelson Exp $ */
 /*
  * ircII: a new irc client.  I like it.  I hope you will too!
  *
@@ -6,7 +6,7 @@
  * Copyright (c) 1991, 1992 Troy Rollo.
  * Copyright (c) 1992-1996 Matthew Green.
  * Copyright  1994 Jake Khuon.
- * Copyright  1993, 2003 EPIC Software Labs.
+ * Copyright  1993, 2010 EPIC Software Labs.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -40,19 +40,19 @@
 /*
  * irc_version is what $J returns, its the common-name for the version.
  */
-const char irc_version[] = "EPIC4-2.10.4";
-const char useful_info[] = "epic4 2 10 4";
+const char irc_version[] = "EPIC5-1.1.4";
+const char useful_info[] = "epic5 1 1 4";
 
 /*
  * internal_version is what $V returns, its the integer-id for the
  * version, and corresponds to the date of release, YYYYMMDD.
  */ 
-const char internal_version[] = "20080904";
+const char internal_version[] = "20120804";
 
 /*
  * In theory, this number is incremented for every commit.
  */
-const unsigned long	commit_id = 778;
+const unsigned long	commit_id = 1667;
 
 /*
  * As a way to poke fun at the current rage of naming releases after
@@ -60,7 +60,7 @@ const unsigned long	commit_id = 778;
  * reality, I have decided to start doing that with EPIC.  These names
  * are intentionally and maliciously silly.  Complaints will be ignored.
  */
-const char ridiculous_version_name[] = "Induration";
+const char ridiculous_version_name[] = "Adumbration";
 
 #define __need_putchar_x__
 #include "status.h"
@@ -81,14 +81,14 @@ const char ridiculous_version_name[] = "Induration";
 #include "ircaux.h"
 #include "commands.h"
 #include "window.h"
-#include "history.h"
 #include "exec.h"
 #include "notify.h"
 #include "mail.h"
 #include "timer.h"
 #include "newio.h"
 #include "parse.h"
-#include "notice.h"
+#include "levels.h"
+#include "extlang.h"
 #include <pwd.h>
 
 
@@ -99,18 +99,12 @@ const char ridiculous_version_name[] = "Induration";
 /* The ``DEFAULT'' port used for irc server connections. */
 int		irc_port = IRC_PORT;
 
-/* Set if ircII should usurp flow control, unset if not.  Probably bogus. */
-int		use_flow_control = 1;
-
-/* Set if ircII should turn on IEXTEN, unset to suppress */
-int		use_iexten = -1;
-
 /* 
  * When a numeric is being processed, this holds the negative value
  * of the numeric.  Its negative so if we pass it to do_hook, it can
  * tell its a numeric and not a named ON event.
  */
-int		current_numeric;
+int		current_numeric = -1;
 
 /* Set if the client is not using a termios interface */
 int		dumb_mode = 0;
@@ -121,12 +115,6 @@ int		background = 0;
 /* Set if the client is checking fd 0 for input. (usu. op of "background") */
 int		use_input = 1;
 
-/* The number of WAIT tokens sent out so far */
-int		waiting_out = 0;
-
-/* The number of WAIT tokens returned so far */
-int		waiting_in = 0;
-
 /*
  * Set when an OPER command is sent out, reset when umode +o or 464 reply
  * comes back.  This is *seriously* bogus.
@@ -145,12 +133,6 @@ Timeval		idle_time = { 0, 0 };
 /* Set to the time the client booted up */
 Timeval		start_time;
 
-/* The number of child processes still unreaped. */
-int		child_dead = 0;
-
-/* Set if the current output is from a trusted source */
-int		trusted_output = 1;
-
 /* Set to 0 when you want to suppress all beeps (such as window repaints) */
 int		global_beep_ok = 1;
 
@@ -166,27 +148,18 @@ volatile int	cntl_c_hit = 0;
 /* This is 1 if we are in the foreground process group */
 int		foreground = 1;
 
-/* This is 0 until your ~/.ircrc is loaded */
-int		ircrc_loaded = 0;
-
-/* This is 0 unless you specify the -B command line flag */
-int		load_ircrc_right_away = 0;
-
 /* This is 1 if you want all logging to be inhibited. Dont leave this on! */
 int		inhibit_logging = 0;
 
-/* This is 1 if we're currently loading the "global" script, 0 afterwards */
-int		loading_global = 0;
-
 /* This is reset every time io() is called.  Use this to save calls to time */
 Timeval		now = {0, 0};
 
-/* When you "highlight ignore", this is what is used to embellish */
-char *		highlight_char;
-
 /* Output which is displayed without modification by the user */
 int		privileged_output = 0;
 
+/* Output which should not trigger %F in hidden windows */
+int		do_window_notifies = 1;
+
 /*
  * If set, outbound connections will be bind()ed to the address
  * specified.  if unset, the default address for your host will
@@ -198,43 +171,37 @@ int		privileged_output = 0;
  * set to addresses that do not belong to the current hostname.
  * If that happens, outbound connections will fail, and its not my fault.
  */
+char *		LocalIPv4HostName = NULL;
+char *		LocalIPv6HostName = NULL;
 char *		LocalHostName = NULL;
-ISA *		LocalIPv4Addr = NULL;
-#ifdef INET6
-ISA6 *		LocalIPv6Addr = NULL;
-#endif
 
 int		inbound_line_mangler = 0,
 		outbound_line_mangler = 0;
 
-char		*startup_file = NULL,		/* full path .epicrc file */
-		*epicrc_file = NULL,		/* full path .epicrc file */
-		*ircrc_file = NULL,		/* full path .ircrc file */
+static char	*epicrc_file = NULL,		/* full path .epicrc file */
+		*ircrc_file = NULL;		/* full path .ircrc file */
+char		*startup_file = NULL,		/* Set when epicrc loaded */
 		*my_path = (char *) 0,		/* path to users home dir */
 		*irc_lib = (char *) 0,		/* path to the ircII library */
 		*default_channel = NULL,	/* Channel to join on connect */
 		nickname[NICKNAME_LEN + 1],	/* users nickname */
 		hostname[NAME_LEN + 1],		/* name of current host */
-		realname[REALNAME_LEN + 1],	/* real name of user */
+#if 0
 		username[NAME_LEN + 1],		/* usernameof user */
 		userhost[NAME_LEN + 1],		/* userhost of user */
+#endif
 		*send_umode = NULL,		/* sent umode */
-		*last_notify_nick = (char *) 0,	/* last detected nickname */
-		empty_string[] = "",		/* just an empty string */
+		*last_notify_nick = (char *) 0;	/* last detected nickname */
+const char	empty_string[] = "",		/* just an empty string */
 		space[] = " ",			/* just a lonely space */
 		on[] = "ON",
-		off[] = "OFF",
+/*		my_off[] = "OFF", */
 		zero[] = "0",
 		one[] = "1",
 		star[] = "*",
 		dot[] = ".",
-		comma[] = ",",
-		*cut_buffer = (char *) 0;	/* global cut_buffer */
-
-fd_set		readables, held_readables;
-fd_set		writables, held_writables;
-int		global_max_fd = -1;
-
+		comma[] = ",";
+unsigned char	*cut_buffer = (unsigned char *) 0;	/* global cut_buffer */
 
 static		char	switch_help[] =
 "Usage: epic [switches] [nickname] [server list]                      \n\
@@ -247,13 +214,10 @@ static		char	switch_help[] =
 #endif
 "      -B\tLoads your .ircrc file before you connect to a server.      \n\
       -d\tThe program should run in ``dumb mode'' (no fancy screen)   \n\
-      -f\tThe program won't mess with your flow control                \n\
-      -F\tThe program will mess with your flow control                \n\
-      -o\tThe program wll turn on IEXTEN terminal setting (^V/^O)     \n\
-      -O\tThe program will turn off IEXTEN (so you can bind ^V/^O)    \n\
       -h\tPrint this help message                                     \n\
       -q\tThe program will not load your .ircrc file                  \n\
       -s\tThe program will not connect to a server upon startup       \n\
+      -S\tEach argument will be tokenised by whitechar                \n\
       -v\tPrint the version of this irc client and exit               \n\
       -x\tRun the client in full X_DEBUG mode                         \n\
       -c <chan>\tJoin <chan> after first connection to a server       \n\
@@ -275,15 +239,9 @@ static SIGNAL_HANDLER(sig_irc_exit)
 void	irc_exit (int really_quit, const char *format, ...)
 {
 	char 	buffer[BIG_BUFFER_SIZE];
-	char *	sub_format;
+	char *	quit_message = NULL;
 	int	old_window_display = window_display;
 	int	value;
-#ifdef PERL
-	extern void perlstartstop(int);
-#endif
-#ifdef TCL
-	extern void tclstartstop(int);
-#endif
 
 	/*
 	 * If we get called recursively, something is hosed.
@@ -308,33 +266,34 @@ void	irc_exit (int really_quit, const char *format, ...)
 	{
 		va_list arglist;
 		va_start(arglist, format);
-		vsnprintf(buffer, BIG_BUFFER_SIZE - 1, format, arglist);
+		vsnprintf(buffer, sizeof(buffer), format, arglist);
 		va_end(arglist);
+		quit_message = buffer;
 	}
 	else
 	{
-		if (!(format = get_string_var(QUIT_MESSAGE_VAR)))
-			format = "%s";
-
-		sub_format = convert_sub_format(format, 's');
-		snprintf(buffer, BIG_BUFFER_SIZE - 1, sub_format, irc_version);
-		new_free(&sub_format);
+		strlcpy(buffer, "Default", sizeof(buffer));
+		quit_message = NULL;
 	}
 
-
 	/* Do some clean up */
 	do_hook(EXIT_LIST, "%s", buffer);
-#ifdef TCL
+#ifdef HAVE_TCL
 	tclstartstop(0);
 #endif
-#ifdef PERL
+#ifdef HAVE_PERL
 	perlstartstop(0);  /* In case there's perl code in the exit hook. */
 #endif
-	close_all_servers(buffer);
+#ifdef HAVE_RUBY
+	ruby_startstop(0);
+#endif
+
+	close_all_servers(quit_message);
 	value = 0;
 	logger(&value);
 	get_child_exit(-1);  /* In case some children died in the exit hook. */
 	clean_up_processes();
+	close_all_dbms();
 
 	/* Arrange to have the cursor on the input line after exit */
 	if (!dumb_mode)
@@ -347,14 +306,16 @@ void	irc_exit (int really_quit, const char *format, ...)
 	
 	/* Try to free as much memory as possible */
 	window_display = 0;
-	dumpcmd(NULL, NULL, NULL);
-	set_lastlog_size(&value);
-	set_history_size(&value);
+/*	dumpcmd(NULL, NULL, NULL); */
 	remove_channel(NULL, 0);
-	destroy_call_stack();
-	remove_bindings();
+	set_lastlog_size(&value);
 	delete_all_windows();
 	destroy_server_list();
+
+	destroy_call_stack();
+	remove_bindings();
+	flush_on_hooks();
+	flush_all_symbols();
 	window_display = old_window_display;
 	fprintf(stdout, "\r");
 	fflush(stdout);
@@ -396,7 +357,7 @@ static SIGNAL_HANDLER(coredump)
 									\n\
 									\n\
 * * * * * * * * * * * * * * * * * * * * * * * *				\n\
-IRC-II has trapped a critical protection error.				\n\
+EPIC has trapped a critical protection error.				\n\
 This is probably due to a bug in the program.				\n\
 									\n\
 If you have access to the 'BUG_FORM' in the ircII source distribution,	\n\
@@ -408,14 +369,16 @@ to it, please dont worry about filling it out.  You might try talking	\n\
 to the person who is in charge of IRC at your site and see if you can	\n\
 get them to help you.							\n\
 									\n\
-This version of IRC II is  --->[%s (%lu)]				\n\
-The date of release is     --->[%s]					\n\
+This version of EPIC is --->[%s (%lu)]					\n\
+The date of release is  --->[%s]					\n\
 									\n\
 * * * * * * * * * * * * * * * * * * * * * * * *				\n\
 The program will now terminate.						\n", irc_version, commit_id, internal_version);
 
 		fflush(stdout);
 		panic_dump_call_stack();
+
+		while ((x_debug & DEBUG_CRASH) && !sleep(1)) {};
 	}
 
         if (x_debug & DEBUG_CRASH)
@@ -441,7 +404,7 @@ static void quit_response (char *dummy, char *ptr)
 }
 
 /* irc_quit: prompts the user if they wish to exit, then does the right thing */
-void irc_quit (char unused, char *not_used)
+BUILT_IN_KEYBINDING(irc_quit)
 {
 	static	int in_it = 0;
 
@@ -465,7 +428,7 @@ static SIGNAL_HANDLER(cntl_c)
 	if (cntl_c_hit++ >= 4)
 		irc_exit(1, "User pressed ^C five times.");
 	else if (cntl_c_hit > 1)
-		kill(getpid(), SIGALRM);
+		kill(getpid(), SIGUSR2);
 }
 
 static SIGNAL_HANDLER(nothing)
@@ -480,9 +443,15 @@ static SIGNAL_HANDLER(sig_user1)
 	clean_up_processes();
 }
 
+static SIGNAL_HANDLER(sig_user2)
+{
+	system_exception++;
+}
+
 static	void	show_version (void)
 {
-	printf("ircII %s (Commit id: %lu) (Date of release: %s)\n\r", irc_version, commit_id, internal_version);
+	printf("ircII %s (Commit id: %lu) (Date of release: %s)\n\r", 
+			irc_version, commit_id, internal_version);
 	exit (0);
 }
 
@@ -506,19 +475,23 @@ static	void	show_version (void)
 static	void	parse_args (int argc, char **argv)
 {
 	int ch;
-	int add_servers = 0;
+	int append_servers = 0;
 	struct passwd *entry;
 	char *ptr = (char *) 0;
 	char *tmp_hostname = NULL;
 	char *the_path = NULL;
 	char *translation_path = NULL;
+	
+	int altargc = 0;
+	char **altargv;
 
 	extern char *optarg;
 	extern int optind;
 
 	*nickname = 0;
-	*realname = 0;
+#if 0
 	*username = 0;
+#endif
 
 	/* 
 	 * Its probably better to parse the environment variables
@@ -531,11 +504,11 @@ static	void	parse_args (int argc, char **argv)
 		{
 			if ((ptr = strchr(entry->pw_gecos, ',')))
 				*ptr = 0;
-			strlcpy(realname, entry->pw_gecos, sizeof realname);
+			set_var_value(DEFAULT_REALNAME_VAR, entry->pw_gecos, 0);
 		}
 
 		if (entry->pw_name && *(entry->pw_name))
-			strlcpy(username, entry->pw_name, sizeof username);
+			set_var_value(DEFAULT_USERNAME_VAR, entry->pw_name, 0);
 
 		if (entry->pw_dir && *(entry->pw_dir))
 			malloc_strcpy(&my_path, entry->pw_dir);
@@ -555,43 +528,49 @@ static	void	parse_args (int argc, char **argv)
 	 * username.  I think everyone would have to agree this is the most
 	 * reasonable way to handle this.
 	 */
-	if (!*username)
+	if (empty(get_string_var(DEFAULT_USERNAME_VAR)))
 		if ((ptr = getenv("LOGNAME")) && *ptr)
-			strlcpy(username, ptr, sizeof username);
+			set_var_value(DEFAULT_USERNAME_VAR, ptr, 0);
 
 #ifndef ALLOW_USER_SPECIFIED_LOGIN
-	if (!*username)
+	if (empty(get_string_var(DEFAULT_USERNAME_VAR)))
 #endif
 		if ((ptr = getenv("IRCUSER")) && *ptr) 
-			strlcpy(username, ptr, sizeof username);
+			set_var_value(DEFAULT_USERNAME_VAR, ptr, 0);
 #ifdef ALLOW_USER_SPECIFIED_LOGIN
-		else if (*username)
+		else if (empty(get_string_var(DEFAULT_USERNAME_VAR)))
 			;
 #endif
 		else if ((ptr = getenv("USER")) && *ptr) 
-			strlcpy(username, ptr, sizeof username);
+			set_var_value(DEFAULT_USERNAME_VAR, ptr, 0);
 		else if ((ptr = getenv("HOME")) && *ptr)
 		{
 			char *ptr2 = strrchr(ptr, '/');
 			if (ptr2)
-				strlcpy(username, ptr2, sizeof username);
+				set_var_value(DEFAULT_USERNAME_VAR, ptr2, 0);
 			else
-				strlcpy(username, ptr, sizeof username);
+				set_var_value(DEFAULT_USERNAME_VAR, ptr, 0);
 		}
 		else
 		{
 			fprintf(stderr, "I dont know what your user name is.\n");
 			fprintf(stderr, "Set your LOGNAME environment variable\n");
-			fprintf(stderr, "and restart IRC II.\n");
+			fprintf(stderr, "and restart EPIC.\n");
 			exit(1);
 		}
 
 	if ((ptr = getenv("IRCNAME")))
-		strlcpy(realname, ptr, sizeof realname);
+		set_var_value(DEFAULT_REALNAME_VAR, ptr, 0);
 	else if ((ptr = getenv("NAME")))
-		strlcpy(realname, ptr, sizeof realname);
-	else if (!*realname)
-		strlcpy(realname, "*Unknown*", sizeof realname);
+		set_var_value(DEFAULT_REALNAME_VAR, ptr, 0);
+	else if ((ptr = getenv("REALNAME")))
+		set_var_value(DEFAULT_REALNAME_VAR, ptr, 0);
+	else
+	{
+		ptr = get_string_var(DEFAULT_REALNAME_VAR);
+		if (!ptr || !*ptr)
+			set_var_value(DEFAULT_REALNAME_VAR, "*Unknown*", 0);
+	}
 
 	if ((ptr = getenv("HOME")))
 		malloc_strcpy(&my_path, ptr);
@@ -626,7 +605,7 @@ static	void	parse_args (int argc, char **argv)
 	else
 		the_path = malloc_sprintf(NULL, DEFAULT_IRCPATH, irc_lib);
 
-	set_string_var(LOAD_PATH_VAR, the_path);
+	set_var_value(LOAD_PATH_VAR, the_path, 0);
 	new_free(&the_path);
 
 	if ((ptr = getenv("IRCHOST")) && *ptr)
@@ -637,13 +616,65 @@ static	void	parse_args (int argc, char **argv)
 	else
 		translation_path = malloc_strdup2(IRCLIB, "/translation/");
 
-	set_string_var(TRANSLATION_PATH_VAR, translation_path);
+	set_var_value(TRANSLATION_PATH_VAR, translation_path, 0);
 	new_free(&translation_path);
+	
+
+	/* The -S-option  / shebang tokeniser */
+	if (argc > 1 && ((argv[1][0] == '-' && argv[1][1] == 'S')
+		|| strchr(argv[1], ' ') != NULL 
+		|| strchr(argv[1], '\t') != NULL)
+	)
+	{
+		int argn, argpos, bufpos, c = 0, n;
+		altargv = new_malloc(sizeof(char *));
+		altargv[0] = new_malloc(strlen(argv[0]) +1);
+		strcpy(altargv[0], argv[0]);
+		altargc = 1;
+		
+		for (argn = 1; argn < argc; argn++)
+		{
+		    int arglen;
+		    char *buffer;
+
+		    arglen =  strlen(argv[argn]);
+		    buffer = alloca(arglen + 1);
+		    argpos = -1;
+		    do
+		    {
+		 	for (n = 0; n <= arglen; ) buffer[n++] = '\0';
+			bufpos = 0;
+			for (argpos++; argpos <= arglen; argpos++)
+			{
+				c = argv[argn][argpos];
+				if (c == '\0' || c == ' ' || c == '\t')
+					break;
+				buffer[bufpos++] = c;
+			}
+			if (bufpos == 0)
+			{
+				if (c == '\0')
+					break;
+				continue;
+			}
+			RESIZE(altargv, char *, altargc +1);
+			altargv[altargc] = new_malloc(bufpos +1);
+			strcpy(altargv[altargc++], buffer);
+			if (c != '\0')
+				continue;
+			break;   	
+		    } while(1);
+		}
+		RESIZE(altargv, char *, altargc +1);
+		altargv[altargc] = NULL;
+		argc = altargc;
+		argv = altargv;
+	}
 
 	/*
 	 * Parse the command line arguments.
 	 */
-	while ((ch = getopt(argc, argv, "aBbc:dfFhH:l:L:n:oOp:qsvxz:")) != EOF)
+	while ((ch = getopt(argc, argv, "aBbc:dhH:l:L:n:p:qsSvxz:")) != EOF)
 	{
 		switch (ch)
 		{
@@ -655,22 +686,6 @@ static	void	parse_args (int argc, char **argv)
 				irc_port = my_atol(optarg);
 				break;
 
-			case 'f': /* Use flow control */
-				use_flow_control = 1;
-				break;
-
-			case 'F': /* dont use flow control */
-				use_flow_control = 0;
-				break;
-
-			case 'o': /* Use IEXTEN */
-				use_iexten = 1;
-				break;
-
-			case 'O': /* dont use IEXTEN */
-				use_iexten = 0;
-				break;
-
 			case 'd': /* use dumb mode */
 				dumb_mode = 1;
 				break;
@@ -680,8 +695,8 @@ static	void	parse_args (int argc, char **argv)
 				malloc_strcpy(&epicrc_file, optarg);
 				break;
 
-			case 'a': /* add server, not replace */
-				add_servers = 1;
+			case 'a': /* append server, not replace */
+				append_servers = 1;
 				break;
 
 			case 'q': /* quick startup -- no .ircrc */
@@ -691,6 +706,9 @@ static	void	parse_args (int argc, char **argv)
 			case 's': /* dont connect - let user choose server */
 				dont_connect = 1;
 				break;
+			
+			case 'S': /* dummy option - to not choke on -S */
+				break;
 
 			case 'b':
 /* siiiiiiiigh */
@@ -713,12 +731,12 @@ static	void	parse_args (int argc, char **argv)
 
 			case 'z':
 #ifdef ALLOW_USER_SPECIFIED_LOGIN
-				strlcpy(username, optarg, sizeof username);
+				set_var_value(DEFAULT_USERNAME_VAR, optarg, 0);
 #endif
 				break;
 
 			case 'B':
-				load_ircrc_right_away = 1;
+				/* Historical option */
 				break;
 
 			case 'c':
@@ -747,11 +765,12 @@ static	void	parse_args (int argc, char **argv)
 	 * so do a final check on whatever nickname we're going to use.
 	 */
 	if (!*nickname)
-		strlcpy(nickname, username, sizeof nickname);
+		strlcpy(nickname, get_string_var(DEFAULT_USERNAME_VAR), 
+				sizeof nickname);
 
 	for (; *argv; argc--, argv++)
 		if (**argv)
-			build_server_list(*argv, NULL);
+			add_servers(*argv, NULL);
 
 	if (!use_input && quick_startup)
 	{
@@ -764,10 +783,10 @@ static	void	parse_args (int argc, char **argv)
 		exit(1);
 	}
 
-	if (strcmp(nickname, "0") && !check_nickname(nickname, 1))
+	if (!check_nickname(nickname, 1))
 	{
 		fprintf(stderr, "Invalid nickname: [%s]\n", nickname);
-		fprintf(stderr, "Please restart IRC II with a valid nickname\n");
+		fprintf(stderr, "Please restart EPIC with a valid nickname\n");
 		exit(1);
 	}
 
@@ -775,15 +794,15 @@ static	void	parse_args (int argc, char **argv)
 	 * Find and build the server lists...
 	 */
 	if ((ptr = getenv("IRCSERVER")))
-		build_server_list(ptr, NULL);
+		add_servers(ptr, NULL);
 
-	if (!server_list_size() || add_servers)
+	if (!server_list_size() || append_servers)
 	{
-		read_server_file();
+		read_default_server_file();
 		if (!server_list_size())
 		{
 			ptr = malloc_strdup(DEFAULT_SERVER);
-			build_server_list(ptr, NULL);
+			add_servers(ptr, NULL);
 			new_free(&ptr);
 		}
 	}
@@ -811,19 +830,35 @@ static	void	parse_args (int argc, char **argv)
 		}
 	}
 
+	if (altargc > 0)
+	{
+		int argn;
+		for (argn = 0; argn < altargc; argn++)
+		{
+			new_free(&(altargv[altargc]));
+		}
+		new_free(&altargv);
+		altargc = 0;
+	}
 	
 	return;
 }
 
 /* fire scripted signal events -pegasus */
-void do_signals(void)
+static void	do_signals(void)
 {
 	int sig_no;
 
 	signals_caught[0] = 0;
 	for (sig_no = 0; sig_no < NSIG; sig_no++)
+	{
 		while (signals_caught[sig_no])
-			do_hook(SIGNAL_LIST, "%d %d", sig_no, signals_caught[sig_no]--);
+		{
+			do_hook(SIGNAL_LIST, "%d %d", sig_no, 
+					signals_caught[sig_no]);
+			signals_caught[sig_no]--;
+		}
+	}
 }
 
 /* 
@@ -847,11 +882,7 @@ static	const	char	*caller[51] = { NULL }; /* XXXX */
 static	int		level = 0,
 			old_level = 0,
 			last_warn = 0;
-static 	const Timeval	right_away = { 0, 0 };
-
 	Timeval		timer;
-	fd_set		rd, wd;
-
 
 	level++;
 	get_time(&now);
@@ -861,59 +892,41 @@ static 	const Timeval	right_away = { 0, 0 };
 
 	if (x_debug & DEBUG_WAITS)
 	{
-		if (level != old_level)
-		{
-			yell("Moving from io level [%d] to level [%d] from [%s]", old_level, level, what);
-			old_level = level;
-		}
+	    if (level != old_level)
+	    {
+		yell("Moving from io level [%d] to level [%d] from [%s]", 
+				old_level, level, what);
+		old_level = level;
+	    }
 	}
 
+	/* Try to avoid letting recursion of io() to get out of control */
 	if (level && (level - last_warn == 5))
 	{
-		last_warn = level;
-		yell("io's recursion level is [%d],  [%s]<-[%s]<-[%s]<-[%s]<-[%s]", level, what, caller[level-1], caller[level-2], caller[level-3], caller[level-4]);
-		if (level % 50 == 0)
-			panic("Ahoy there matey!  Abandon ship!");
+	    last_warn = level;
+	    yell("io's recursion level is [%d],  [%s]<-[%s]<-[%s]<-[%s]<-[%s]",
+			level, what, caller[level-1], caller[level-2], 
+					caller[level-3], caller[level-4]);
+	    if (level % 50 == 0)
+		panic(1, "Ahoy there matey!  Abandon ship!");
 	}
 	else if (level && (last_warn - level == 5))
-		last_warn -= 5;
-
+	    last_warn -= 5;
 
 	caller[level] = what;
 
-#if 1
-	/* 
-	 * XXXXXXX! It pains me greatly to call this here.
-	 * This is an experimental test to see if i can get
-	 * away with not having to call update_all_windows()
-	 * in about a zillion other places by doing it here.
-	 */
-	update_all_windows();
-#endif
-
-	/* SET UP FD SETS */
-	rd = readables;
-	wd = writables;
-
-	/* If there is a timer that expires sooner, wait for that */
-	/* There is now a timer at all times, so this is our baseline */
+	/* Calculate the time to the next timer timeout */
 	timer = TimerTimeout();
 
-	/* If for any reason the timeout is negative, do a poll */
-	if (time_diff(right_away, timer) < 0)
-		timer = right_away;
-
 	/* GO AHEAD AND WAIT FOR SOME DATA TO COME IN */
-	switch (new_select(&rd, &wd, &timer))
+	make_window_current(NULL);
+	switch (do_wait(&timer))
 	{
-		/* Timeout -- nothing interesting. */
+		/* Timeout -- Need to do timers */
 		case 0:
 		{
 			get_time(&now);
-#ifdef HAVE_SSL
-			/* Yes, this is slow, but we have to check for this */
-			do_server(&rd, &wd);
-#endif
+			ExecuteTimers();
 			break;
 		}
 
@@ -935,38 +948,58 @@ static 	const Timeval	right_away = { 0, 0 };
 		default:
 		{
 			get_time(&now);
-			make_window_current(NULL);
-			dcc_check(&rd, &wd);
-			do_server(&rd, &wd);
-			do_processes(&rd, &wd);
-			do_screens(&rd, &wd);
+			do_filedesc();
 			break;
 		} 
 	}
 
+	/* 
+	 * Various things that need to be done synchronously...
+	 */
+
+	/* deal with caught signals - pegasus */
 	if (signals_caught[0] != 0)
 		do_signals();
-	ExecuteTimers();
+
+	/* Account for dead child processes */
 	get_child_exit(-1);
+
+	/* Run /DEFERed commands */
 	if (level == 1 && need_defered_commands)
 		do_defered_commands();
 
+	/* Make sure all the servers are connected that ought to be */
+	window_check_servers();
+
+	/* Make sure all the channels are joined that ought to be */
+	window_check_channels();
+
+	/* Redraw the screen after a SIGCONT */
+	if (need_redraw)
+		redraw_all_screens();
+
+	/* Make sure all the windows and status bars are made current */
+	update_all_windows();
+
+	/* Move the cursor back to the input line */
 	cursor_to_input();
 
-	/* (set in term.c) -- we need to redraw the screen */
-	if (need_redraw)
-		refresh_a_screen(main_screen);
-
-	window_check_channels();
-	update_all_windows();
+	/* Reclaim any alloca()ed space */
 	alloca(0);
+
+	/* Release this io() accounting level */
 	caller[level] = NULL;
 	level--;
 
 #ifdef DELAYED_FREES
+	/* Reclaim any malloc()ed space */
 	if (level == 0 && need_delayed_free)
 		do_delayed_frees();
 #endif
+
+	if (level == 0)
+		check_message_from_queue();
+
 	return;
 }
 
@@ -1118,9 +1151,26 @@ static void check_invalid_host (void)
 	return;
 }
 
+/*************************************************************************/
+void    load_ircrc (void)
+{
+	if (startup_file || quick_startup)
+		return;
+
+	if (access(epicrc_file, R_OK) == 0)
+		startup_file = malloc_strdup(epicrc_file);
+	else if (access(ircrc_file, R_OK) == 0)
+		startup_file = malloc_strdup(ircrc_file);
+	else
+		startup_file = malloc_strdup("global");
+
+	load("LOAD", startup_file, empty_string);
+}
+
 /*************************************************************************/
 int 	main (int argc, char *argv[])
 {
+	setvbuf(stdout, NULL, _IOLBF, 1024);
 #ifdef SOCKS
 	SOCKSinit(argv[0]);
 #endif
@@ -1128,37 +1178,60 @@ int 	main (int argc, char *argv[])
 	check_password();
 	check_valid_user();
 	check_invalid_host();
+
+#ifdef WITH_THREADED_STDOUT
+	tio_init();
+#endif
+
+	init_levels();
+	init_transforms();
+	init_variables_stage1();
 	parse_args(argc, argv);
 	init_binds();
 	init_keys();
+	init_commands();
+	init_functions();
+	init_expandos();
+	init_newio();
 
-	fprintf(stderr, "EPIC Version 4 -- %s\n", ridiculous_version_name);
-	fprintf(stderr, "EPIC Software Labs (2004)\n");
-	fprintf(stderr, "Version (%s), Commit Id (%lu) -- Date (%s)\n", irc_version, commit_id, internal_version);
+	fprintf(stderr, "EPIC Version 5 -- %s\n", ridiculous_version_name);
+	fprintf(stderr, "EPIC Software Labs (2006)\n");
+	fprintf(stderr, "Version (%s), Commit Id (%lu) -- Date (%s)\n", 
+				irc_version, commit_id, internal_version);
 	fprintf(stderr, "%s\n", compile_info);
-	fprintf(stderr, "Process [%d]", getpid());
-	if (isatty(0))
-		fprintf(stderr, " connected to tty [%s]", ttyname(0));
-	else
-		dumb_mode = 1;
-	fprintf(stderr, "\n");
-
-	FD_ZERO(&readables);
-	FD_ZERO(&writables);
-	FD_ZERO(&held_readables);
-	FD_ZERO(&held_writables);
 
+#ifndef NO_JOB_CONTROL
 	/* If we're a bot, do the bot thing. */
-	if (!use_input && fork())
-		_exit(0);
+	if (!use_input)
+	{
+		pid_t	child_pid;
+
+		child_pid = fork();
+		if (child_pid == -1)
+		{
+			fprintf(stderr, "Could not fork a child process: %s\n",
+					strerror(errno));
+			_exit(1);
+		}
+		else if (child_pid > 0)
+		{
+			fprintf(stderr, "Process [%d] running in background\n",
+					child_pid);
+			_exit(0);
+		}
+	}
+	else
+#endif
+	{
+		fprintf(stderr, "Process [%d]", getpid());
+		if (isatty(0))
+			fprintf(stderr, " connected to tty [%s]", ttyname(0));
+		else
+			dumb_mode = 1;
+		fprintf(stderr, "\n");
+	}
 
-	/* make sure we don't start with spurious signals events firing */
-	memset(&signals_caught, 0, NSIG * sizeof(int));
-	/* hook all signals! */
 	init_signals();
-	/* we *might* want to check for SIG_ERR from the above function.
-	 * i leave it to hop to decide what to do on SIG_ERR. -pegasus 
-	 */
 
 	/* these should be taken by both dumb and smart displays */
 	my_signal(SIGSEGV, coredump);
@@ -1171,14 +1244,22 @@ int 	main (int argc, char *argv[])
 	my_signal(SIGINT, cntl_c);
 	my_signal(SIGALRM, nothing);
 	my_signal(SIGUSR1, sig_user1);
+	my_signal(SIGUSR2, sig_user2);
 
+	message_from(NULL, LEVEL_OTHER);
+
+	/* 
+	 * We use dumb mode for -d, -b, when stdout is redirected to a file,
+	 * or as a failover if init_screen() fails. 
+	 */
 	if ((dumb_mode == 0) && (init_screen() == 0))
 	{
 		my_signal(SIGCONT, term_cont);
 		my_signal(SIGWINCH, sig_refresh_screen);
-		init_variables();
+		init_variables_stage2();
 		build_status(NULL);
-		update_input(UPDATE_ALL);
+		permit_status_update(1);
+		update_input(NULL, UPDATE_ALL);
 	}
 	else
 	{
@@ -1190,7 +1271,7 @@ int 	main (int argc, char *argv[])
 		dumb_mode = 1;		/* Just in case */
 		create_new_screen();
 		new_window(main_screen);
-		init_variables();
+		init_variables_stage2();
 		build_status(NULL);
 	}
 
@@ -1200,17 +1281,17 @@ int 	main (int argc, char *argv[])
 	/* The all-collecting stack frame */
 	make_local_stack("TOP");
 
-	/* XXXX Move this somewhere else eventually XXXX */
-	if (load_ircrc_right_away)
-		load_ircrc();
+	load_ircrc();
 
 	set_input(empty_string);
-	set_input_prompt(get_string_var(INPUT_PROMPT_VAR));
 
 	if (dont_connect)
 		display_server_list();		/* Let user choose server */
 	else
-		reconnect(NOSERV, 0);		/* Connect to default server */
+	{
+		current_window->server = 0;	/* Connect to default server */
+		window_check_servers();
+	}
 
 	get_time(&idle_time);
 	reset_system_timers();
