@@ -1,4 +1,4 @@
-/* $EPIC: status.c,v 1.35 2003/12/07 20:16:52 jnelson Exp $ */
+/* $EPIC: status.c,v 1.75 2012/03/25 02:27:46 jnelson Exp $ */
 /*
  * status.c: handles the status line updating, etc for IRCII 
  *
@@ -38,7 +38,7 @@
 #define __need_putchar_x__
 #include "irc.h"
 #include "dcc.h"
-#include "term.h"
+#include "termx.h"
 #include "status.h"
 #include "server.h"
 #include "vars.h"
@@ -63,7 +63,7 @@
  * Maximum number of "%" expressions in a status line format.  If you change
  * this number, you must manually change the snprintf() in make_status 
  */
-#define STATUS_FUNCTION(x) Char * x (Window *window, int map, int key)
+#define STATUS_FUNCTION(x) Char * x (Window *window, short map, char key)
 #define MAX_FUNCTIONS 40
 #define MAX_STATUS_USER 39
 
@@ -74,6 +74,7 @@ STATUS_FUNCTION(status_chanop);
 STATUS_FUNCTION(status_ssl);
 STATUS_FUNCTION(status_channel);
 STATUS_FUNCTION(status_server);
+STATUS_FUNCTION(status_network);
 STATUS_FUNCTION(status_mode);
 STATUS_FUNCTION(status_umode);
 STATUS_FUNCTION(status_insert_mode);
@@ -84,6 +85,7 @@ STATUS_FUNCTION(status_user);
 STATUS_FUNCTION(status_dcc);
 STATUS_FUNCTION(status_dcc_all);
 STATUS_FUNCTION(status_hold);
+STATUS_FUNCTION(status_holdmode);
 STATUS_FUNCTION(status_version);
 STATUS_FUNCTION(status_clock);
 STATUS_FUNCTION(status_hold_lines);
@@ -102,6 +104,7 @@ STATUS_FUNCTION(status_windowspec);
 STATUS_FUNCTION(status_percent);
 STATUS_FUNCTION(status_test);
 STATUS_FUNCTION(status_swappable);
+STATUS_FUNCTION(status_activity);
 
 /* These are used as placeholders for some expandos */
 static	char	*mode_format 		= (char *) 0;
@@ -133,94 +136,102 @@ int     status_updates_permitted = 0;
  * specifying the map.
  */
 struct status_formats {
-	int	map;
+	short	map;
 	char 	key;
-	Char	*(*callback_function)(Window *, int, int);
+	Char	*(*callback_function)(Window *, short, char);
 	char	**format_var;
-	int	format_set;
+	int	*format_set;
 };
 struct status_formats status_expandos[] = {
-{ 0, 'A', status_away,          NULL, 			-1 },
-{ 0, 'B', status_hold_lines,    &hold_lines_format,	STATUS_HOLD_LINES_VAR },
-{ 0, 'C', status_channel,       &channel_format,	STATUS_CHANNEL_VAR },
-{ 0, 'D', status_dcc, 	        NULL, 			-1 },
-{ 0, 'F', status_notify_windows,&notify_format,		STATUS_NOTIFY_VAR },
-{ 0, 'H', status_hold,		NULL,			-1 },
-{ 0, 'I', status_insert_mode,   NULL,			-1 },
-{ 0, 'K', status_scrollback,	NULL,			-1 },
-{ 0, 'L', status_cpu_saver_mode,&cpu_saver_format,	STATUS_CPU_SAVER_VAR },
-{ 0, 'M', status_mail,		&mail_format,		STATUS_MAIL_VAR },
-{ 0, 'N', status_nickname,	&nick_format,		STATUS_NICK_VAR },
-{ 0, 'O', status_overwrite_mode,NULL,			-1 },
-{ 0, 'P', status_position,      NULL,			-1 },
-{ 0, 'Q', status_query_nick,    &query_format,		STATUS_QUERY_VAR },
-{ 0, 'R', status_refnum,        NULL, 			-1 },
-{ 0, 'S', status_server,        &server_format,     	STATUS_SERVER_VAR },
-{ 0, 'T', status_clock,         &clock_format,      	STATUS_CLOCK_VAR },
-{ 0, 'U', status_user,		NULL, 			-1 },
-{ 0, 'V', status_version,	NULL, 			-1 },
-{ 0, 'W', status_window,	NULL, 			-1 },
-{ 0, 'X', status_user,		NULL, 			-1 },
-{ 0, 'Y', status_user,		NULL, 			-1 },
-{ 0, 'Z', status_user,		NULL, 			-1 },
-{ 0, '#', status_umode,		&umode_format,	     	STATUS_UMODE_VAR },
-{ 0, '%', status_percent,	NULL, 			-1 },
-{ 0, '*', status_oper,		NULL, 			-1 },
-{ 0, '+', status_mode,		&mode_format,       	STATUS_MODE_VAR },
-{ 0, '.', status_windowspec,	NULL, 			-1 },
-{ 0, '=', status_voice,		NULL, 			-1 },
-{ 0, '>', status_right_justify,	NULL, 			-1 },
-{ 0, '@', status_chanop,	NULL, 			-1 },
-{ 0, '|', status_ssl,		NULL,			-1 },
-{ 0, '0', status_user,		NULL, 			-1 },
-{ 0, '1', status_user,		NULL, 			-1 },
-{ 0, '2', status_user,		NULL, 			-1 },
-{ 0, '3', status_user,		NULL, 			-1 },
-{ 0, '4', status_user,		NULL, 			-1 },
-{ 0, '5', status_user,		NULL, 			-1 },
-{ 0, '6', status_user,		NULL, 			-1 },
-{ 0, '7', status_user,		NULL, 			-1 },
-{ 0, '8', status_user,		NULL, 			-1 },
-{ 0, '9', status_user,		NULL, 			-1 },
-{ 1, '0', status_user,		NULL, 			-1 },
-{ 1, '1', status_user,		NULL, 			-1 },
-{ 1, '2', status_user,		NULL, 			-1 },
-{ 1, '3', status_user,		NULL, 			-1 },
-{ 1, '4', status_user,		NULL, 			-1 },
-{ 1, '5', status_user,		NULL, 			-1 },
-{ 1, '6', status_user,		NULL, 			-1 },
-{ 1, '7', status_user,		NULL, 			-1 },
-{ 1, '8', status_user,		NULL, 			-1 },
-{ 1, '9', status_user,		NULL, 			-1 },
-{ 1, 'D', status_dcc_all,	NULL, 			-1 },
-{ 1, 'F', status_notify_windows,&notify_format,		STATUS_NOTIFY_VAR },
-{ 1, 'K', status_scroll_info,	NULL,			-1 },
-{ 1, 'R', status_refnum_real,   NULL, 			-1 },
-{ 1, 'S', status_server,        &server_format,     	STATUS_SERVER_VAR },
-{ 1, 'T', status_test,		NULL,			-1 },
-{ 1, 'W', status_swappable,	NULL,			-1 },
-{ 1, '+', status_mode,		&mode_format,       	STATUS_MODE_VAR },
-{ 2, '0', status_user,	 	NULL, 			-1 },
-{ 2, '1', status_user,	 	NULL, 			-1 },
-{ 2, '2', status_user,	 	NULL, 			-1 },
-{ 2, '3', status_user,	 	NULL, 			-1 },
-{ 2, '4', status_user,		NULL, 			-1 },
-{ 2, '5', status_user,	 	NULL, 			-1 },
-{ 2, '6', status_user,	 	NULL, 			-1 },
-{ 2, '7', status_user,	 	NULL,			-1 },
-{ 2, '8', status_user,	 	NULL, 			-1 },
-{ 2, '9', status_user,	 	NULL, 			-1 },
-{ 2, 'S', status_server,        &server_format,     	STATUS_SERVER_VAR },
-{ 3, '0', status_user,	 	NULL, 			-1 },
-{ 3, '1', status_user,	 	NULL, 			-1 },
-{ 3, '2', status_user,	 	NULL, 			-1 },
-{ 3, '3', status_user,	 	NULL, 			-1 },
-{ 3, '4', status_user,	 	NULL, 			-1 },
-{ 3, '5', status_user,	 	NULL, 			-1 },
-{ 3, '6', status_user,	 	NULL, 			-1 },
-{ 3, '7', status_user,	 	NULL, 			-1 },
-{ 3, '8', status_user,	 	NULL, 			-1 },
-{ 3, '9', status_user,	 	NULL, 			-1 }
+{ 0, 'A', status_away,          NULL, 			NULL },
+{ 0, 'B', status_hold_lines,    &hold_lines_format,	&STATUS_HOLD_LINES_VAR},
+{ 0, 'C', status_channel,       &channel_format,	&STATUS_CHANNEL_VAR },
+{ 0, 'D', status_dcc, 	        NULL, 			NULL },
+{ 0, 'E', status_activity,	NULL,			NULL },
+{ 0, 'F', status_notify_windows,&notify_format,		&STATUS_NOTIFY_VAR },
+{ 0, 'G', status_network,		NULL,			NULL },
+{ 0, 'H', status_hold,		NULL,			NULL },
+{ 0, 'I', status_insert_mode,   NULL,			NULL },
+{ 0, 'K', status_scrollback,	NULL,			NULL },
+{ 0, 'L', status_cpu_saver_mode,&cpu_saver_format,	&STATUS_CPU_SAVER_VAR },
+{ 0, 'M', status_mail,		&mail_format,		&STATUS_MAIL_VAR },
+{ 0, 'N', status_nickname,	&nick_format,		&STATUS_NICKNAME_VAR },
+{ 0, 'O', status_overwrite_mode,NULL,			NULL },
+{ 0, 'P', status_position,      NULL,			NULL },
+{ 0, 'Q', status_query_nick,    &query_format,		&STATUS_QUERY_VAR },
+{ 0, 'R', status_refnum,        NULL, 			NULL },
+{ 0, 'S', status_server,        &server_format,     	&STATUS_SERVER_VAR },
+{ 0, 'T', status_clock,         &clock_format,      	&STATUS_CLOCK_VAR },
+{ 0, 'U', status_user,		NULL, 			NULL },
+{ 0, 'V', status_version,	NULL, 			NULL },
+{ 0, 'W', status_window,	NULL, 			NULL },
+{ 0, 'X', status_user,		NULL, 			NULL },
+{ 0, 'Y', status_user,		NULL, 			NULL },
+{ 0, 'Z', status_user,		NULL, 			NULL },
+{ 0, '#', status_umode,		&umode_format,	     	&STATUS_UMODE_VAR },
+{ 0, '%', status_percent,	NULL, 			NULL },
+{ 0, '*', status_oper,		NULL, 			NULL },
+{ 0, '+', status_mode,		&mode_format,       	&STATUS_MODE_VAR },
+{ 0, '.', status_windowspec,	NULL, 			NULL },
+{ 0, '=', status_voice,		NULL, 			NULL },
+{ 0, '>', status_right_justify,	NULL, 			NULL },
+{ 0, '@', status_chanop,	NULL, 			NULL },
+{ 0, '|', status_ssl,		NULL,			NULL },
+{ 0, '0', status_user,		NULL, 			NULL },
+{ 0, '1', status_user,		NULL, 			NULL },
+{ 0, '2', status_user,		NULL, 			NULL },
+{ 0, '3', status_user,		NULL, 			NULL },
+{ 0, '4', status_user,		NULL, 			NULL },
+{ 0, '5', status_user,		NULL, 			NULL },
+{ 0, '6', status_user,		NULL, 			NULL },
+{ 0, '7', status_user,		NULL, 			NULL },
+{ 0, '8', status_user,		NULL, 			NULL },
+{ 0, '9', status_user,		NULL, 			NULL },
+{ 1, '0', status_user,		NULL, 			NULL },
+{ 1, '1', status_user,		NULL, 			NULL },
+{ 1, '2', status_user,		NULL, 			NULL },
+{ 1, '3', status_user,		NULL, 			NULL },
+{ 1, '4', status_user,		NULL, 			NULL },
+{ 1, '5', status_user,		NULL, 			NULL },
+{ 1, '6', status_user,		NULL, 			NULL },
+{ 1, '7', status_user,		NULL, 			NULL },
+{ 1, '8', status_user,		NULL, 			NULL },
+{ 1, '9', status_user,		NULL, 			NULL },
+{ 1, 'D', status_dcc_all,	NULL, 			NULL },
+{ 1, 'F', status_notify_windows,&notify_format,		&STATUS_NOTIFY_VAR },
+{ 1, 'H', status_holdmode,	NULL,			NULL },
+{ 1, 'K', status_scroll_info,	NULL,			NULL },
+{ 1, 'R', status_refnum_real,   NULL, 			NULL },
+{ 1, 'S', status_server,        &server_format,     	&STATUS_SERVER_VAR },
+{ 1, 'T', status_test,		NULL,			NULL },
+{ 1, 'W', status_swappable,	NULL,			NULL },
+{ 1, '+', status_mode,		&mode_format,       	&STATUS_MODE_VAR },
+{ 2, '0', status_user,	 	NULL, 			NULL },
+{ 2, '1', status_user,	 	NULL, 			NULL },
+{ 2, '2', status_user,	 	NULL, 			NULL },
+{ 2, '3', status_user,	 	NULL, 			NULL },
+{ 2, '4', status_user,		NULL, 			NULL },
+{ 2, '5', status_user,	 	NULL, 			NULL },
+{ 2, '6', status_user,	 	NULL, 			NULL },
+{ 2, '7', status_user,	 	NULL,			NULL },
+{ 2, '8', status_user,	 	NULL, 			NULL },
+{ 2, '9', status_user,	 	NULL, 			NULL },
+{ 2, 'S', status_server,        &server_format,     	&STATUS_SERVER_VAR },
+{ 2, 'W', status_window,	NULL, 			NULL },
+{ 2, '+', status_mode,		&mode_format,       	&STATUS_MODE_VAR },
+{ 3, '0', status_user,	 	NULL, 			NULL },
+{ 3, '1', status_user,	 	NULL, 			NULL },
+{ 3, '2', status_user,	 	NULL, 			NULL },
+{ 3, '3', status_user,	 	NULL, 			NULL },
+{ 3, '4', status_user,	 	NULL, 			NULL },
+{ 3, '5', status_user,	 	NULL, 			NULL },
+{ 3, '6', status_user,	 	NULL, 			NULL },
+{ 3, '7', status_user,	 	NULL, 			NULL },
+{ 3, '8', status_user,	 	NULL, 			NULL },
+{ 3, '9', status_user,	 	NULL, 			NULL },
+{ 3, 'S', status_server,        &server_format,     	&STATUS_SERVER_VAR },
+{ 3, 'W', status_window,	NULL, 			NULL },
+{ 3, '+', status_mode,		&mode_format,       	&STATUS_MODE_VAR }
 };
 #define NUMBER_OF_EXPANDOS (sizeof(status_expandos) / sizeof(struct status_formats))
 
@@ -289,7 +300,7 @@ static void	build_status_format (Status *s, int k)
 {
 	char	buffer[BIG_BUFFER_SIZE + 1];
 	int	cp;
-	int	map;
+	short	map;
 	char	key;
 	unsigned	i;
 	Char	*raw = s->line[k].raw;
@@ -336,9 +347,9 @@ static void	build_status_format (Status *s, int k)
 
 			if (status_expandos[i].format_var)
 				new_free(status_expandos[i].format_var);
-			if (status_expandos[i].format_set != -1)
+			if (status_expandos[i].format_set != NULL)
 				*(status_expandos[i].format_var) = 
-					convert_sub_format(get_string_var(status_expandos[i].format_set), key);
+					convert_sub_format(get_string_var(*status_expandos[i].format_set), key);
 
 			*format++ = '%';
 			*format++ = 's';
@@ -441,7 +452,7 @@ void	rebuild_a_status (Window *w)
  * This function is a /SET callback, so it must always take a (char *) as an
  * argument even though we don't care about it.
  */
-void	build_status	(const void *stuff)
+void	build_status	(void *stuff)
 {
 	Window 	*w = NULL;
 
@@ -460,7 +471,7 @@ static void	init_status	(void)
 {
 	int	i, k;
 
-	main_status.double_status = 0;
+	main_status.number = 1;
 	main_status.special = 0;
 	for (i = 0; i < 3; i++)
 	{
@@ -496,21 +507,21 @@ int     permit_status_update (int flag)
  */
 int	make_status (Window *window, int must_redraw)
 {
-	int	status_line;
-	u_char	buffer	    [BIG_BUFFER_SIZE + 1];
-	u_char	lhs_buffer  [BIG_BUFFER_SIZE + 1];
-	u_char	rhs_buffer  [BIG_BUFFER_SIZE + 1];
-	Char	*func_value [MAX_FUNCTIONS];
-	u_char	*ptr;
-	size_t	save_size;
+	int		status_line;
+	unsigned char	buffer	    [BIG_BUFFER_SIZE + 1];
+	unsigned char	lhs_buffer  [BIG_BUFFER_SIZE + 1];
+	unsigned char	rhs_buffer  [BIG_BUFFER_SIZE + 1];
+	const char	*func_value [MAX_FUNCTIONS];
+	unsigned char	*ptr;
+	size_t		save_size;
 
 	/* We do NOT redraw status bars for hidden windows */
 	if (!window->screen || !status_updates_permitted)
 		return -1;
 
-	for (status_line = 0; status_line < window->status.double_status + 1; status_line++)
+	for (status_line = 0; status_line < window->status.number; status_line++)
 	{
-		u_char	lhs_fillchar[6],
+	unsigned char	lhs_fillchar[6],
 			rhs_fillchar[6],
 			*fillchar = lhs_fillchar,
 			*lhp = lhs_buffer,
@@ -521,35 +532,25 @@ int	make_status (Window *window, int must_redraw)
 		int	in_rhs = 0,
 			pr_lhs = 0,
 			pr_rhs = 0,
-			line,
+			line = 0,	/* XXX gcc4 lameness */
 			*prc = &pr_lhs, 
 			i;
 
 		fillchar[0] = fillchar[1] = 0;
 
 		/*
-		 * If status line gets to one, then that means that
-		 * window->double_status is not zero.  That means that
-		 * the status line we're working on is STATUS2.
+		 * Figure out which of the three status bars we're creating.
 		 */
-		if (status_line)
-			line = 2;
-
-		/*
-		 * If status_line is zero, and window->double_status is
-		 * not zero (double status line is on) then we're working
-		 * on STATUS1.
-		 */
-		else if (window->status.double_status)
-			line = 1;
-
-		/*
-		 * So status_line is zero and window->double_status is zero.
-		 * So we're working on STATUS (0).
-		 */
-		else
+		if (window->status.number == 1 && status_line == 0)
 			line = 0;
-
+		else if (window->status.number == 2 && status_line == 0)
+			line = 1;
+		else if (window->status.number == 2 && status_line == 1)
+			line = 2;
+		else
+			panic(1, "window->status.number is [%d] and status_line "
+				"is [%d] and that makes no sense!", 
+				window->status.number, status_line);
 
 		/*
 		 * Sanity check:  If the status format doesnt exist, dont do
@@ -570,12 +571,13 @@ int	make_status (Window *window, int must_redraw)
 				return -1;
 
 			if (window->status.line[line].func[i] == NULL)
-				panic("status callback null.  Window [%d], line [%d], function [%d]", window->refnum, line, i);
+				panic(1, "status callback null.  Window [%d], line [%d], function [%d]", window->refnum, line, i);
 			func_value[i] = window->status.line[line].func[i]
 				(window, window->status.line[line].map[i],
 				 window->status.line[line].key[i]);
 		}
 
+#if 0
 		/*
 		 * If the REVERSE_STATUS_LINE var is on, then put a reverse
 		 * character in the first position (itll get translated to
@@ -584,6 +586,7 @@ int	make_status (Window *window, int must_redraw)
 		if (get_int_var(REVERSE_STATUS_LINE_VAR))
 			*buffer = REV_TOG , str = buffer + 1;
 		else
+#endif
 			str = buffer;
 
 		/*
@@ -613,7 +616,6 @@ int	make_status (Window *window, int must_redraw)
 		 */
 		if (get_int_var(STATUS_DOES_EXPANDOS_VAR))
 		{
-			int  af = 0;
 			int  old_fs = from_server;
 			Window *old = current_window;
 			int	owd = window_display;
@@ -621,7 +623,7 @@ int	make_status (Window *window, int must_redraw)
 			current_window = window;
 			from_server = current_window->server;
 			window_display = 0;
-			str = expand_alias(buffer, empty_string, &af, NULL);
+			str = expand_alias(buffer, empty_string);
 			window_display = owd;
 			from_server = old_fs;
 			current_window = old;
@@ -636,7 +638,7 @@ int	make_status (Window *window, int must_redraw)
 		 * with nothing but logical characters, which are then easy
 		 * to count. :-)
 		 */
-		str = normalize_string(buffer, 3);
+		str = new_normalize_string(buffer, 3, display_line_mangler);
 
 		/*
 		 * Count out the characters.
@@ -738,7 +740,7 @@ int	make_status (Window *window, int must_redraw)
 		/*
 		 * No rhs?  If the user wants us to pad it out, do so.
 		 */
-		else if (get_int_var(FULL_STATUS_LINE_VAR))
+		else 
 		{
 			int chars = window->screen->co - pr_lhs - 1;
 
@@ -820,20 +822,107 @@ int	make_status (Window *window, int must_redraw)
 
 
 /*
- * These are the functions that all of the status expandoes invoke
+ * "Ack!  What are these macros?" you may be asking.
+ *
+ * Traditionally, when ircII redraws your status bar, it calls a function
+ * for each status expando you have in your status bar (ie, %T calls 
+ * status_time()).  Each function returns a malloced string, which is then
+ * sprintf()d into a bigger string, and the malloced string is freed.
+ * This is inefficient, because status redraws happen pretty frqeuently
+ * and malloc()ing is very expensive.
+ *
+ * Historically, when epic redraws your status bar, it calls a function
+ * for each status expando, just like ircII, but each function has its
+ * own private 'static' variable used for return values.  This requires
+ * less cpu since there isn't a lot of extra malloc+free stuff going on.
+ * The downside is it's possible to fill/overrun these buffers, which 
+ * would lead to the values being truncated.
+ *
+ * As a compromise, we now use a two-cycle system.  We continue to use the
+ * static variable as we always have, but we malloc() it off only the first
+ * time we use it, and any time it needs to grow in size.  This permits us
+ * to avoid malloc+free() every single time for every single status bar, and
+ * avoid possible truncation/overrun of predefined buffer sizes.
+ *
+ * These macros are only necessary for those status expandos that might
+ * return values of indeterminite length.  It's not necessary for those 
+ * functions where the return value is hardcoded or has known limited size.
+ *
+ * So what we do is do a snprintf() into the return value, and snprintf()
+ * tells us if the result was truncated.  If it was truncated, we resize
+ * the return value buffer and then do the snprintf() again, and this time
+ * it won't be truncated. 
  */
 
+/* 
+ * This macro declares the expandable return value variables.
+ *	my_bufferx	Is the actual resizable return value.  It is malloced
+ *			of when first used, and then resized any time a return
+ *			value would overflow it.
+ *	my_bufferxsize	The actual size we have for the return value. This is 
+ *			the "having" space.
+ *	actual_size	The actual size of what we want to return.  This is the
+ *			"needed" space.
+ *
+ */
+#define STATUS_VARS \
+	static char *my_bufferx = NULL; \
+	static int my_bufferxsize = 64; \
+	int	actual_size;
+
+/* After you change 'my_bufferxspace', call this to update 'my_bufferx' */
+#define CHECK \
+	RESIZE(my_bufferx, char, my_bufferxsize);
+
+/* 
+ * After you change 'actual_size', call this to check if the results
+ * overflowed the return value buffer.  If it did overflow, it resizes
+ * the return value buffer so it won't overflow again.
+ */
+#define RECHECK \
+	if (actual_size < 0) 	/* Die Die Die */ \
+		my_bufferxsize += 16;	\
+	else if (actual_size < my_bufferxsize) \
+		break; \
+	else					\
+		my_bufferxsize = actual_size + 1; \
+	CHECK
+
+/*
+ * Do a sprintf() using the printf format "fmt", which must contain only
+ * one "%s" and no other %-formats, using the 'arg' string.  If the result
+ * overflows the return buffer, the return buffer is resized and we do it
+ * a second time.
+ */
+#define PRESS(fmt, arg) \
+	if (! fmt ) return empty_string; \
+	CHECK \
+	do { \
+		actual_size = snprintf(my_bufferx, my_bufferxsize, fmt, arg); \
+		RECHECK \
+	} while (1);
+
+/*
+ * Return the return buffer.  The reason this is not part of 'PRESS' is 
+ * because sometimes you may want to do something between the PRESS but 
+ * before the RETURN, based on the result.
+ */
+#define RETURN return my_bufferx;
+
+/***********************************************************************/
+/*
+ * These are the functions that all of the status expandoes invoke
+ */
 
 /*
  * This is your current nickname in the window.
  */
 STATUS_FUNCTION(status_nickname)
 {
-static	char	my_buffer[64];
+	STATUS_VARS
 
-	snprintf(my_buffer, 63, nick_format, 
-			get_server_nickname(window->server));
-	return my_buffer;
+	PRESS(nick_format, get_server_nickname(window->server))
+	RETURN
 }
 
 /*
@@ -841,21 +930,15 @@ static	char	my_buffer[64];
  */
 STATUS_FUNCTION(status_server)
 {
-	char	*rest;
-const	char	*n;
-	char	*name;
-	char	*next;
-static	char	my_buffer[64];
-	size_t	len;
+	STATUS_VARS
+const	char	*n = NULL;
 
-#ifdef OLD_STATUS_S_EXPANDO_BEHAVIOR
 	/*
 	 * If there is only one server, dont bother telling the user
 	 * what it is.
 	 */
-	if (connected_to_server == 1 && map == 0)
+	if (map == 0 && connected_to_server == 1)
 		return empty_string;
-#endif
 
 	/*
 	 * If this window isnt connected to a server, say so.
@@ -863,64 +946,35 @@ static	char	my_buffer[64];
 	if (window->server == NOSERV)
 		return "No Server";
 
-	/*
-	 * If the user doesnt want this expando, dont force it.
-	 */
-	if (!server_format)
-		return empty_string;
+	/* Map 0 uses the shortname, shown when multi-connected */
+	/* Map 1 uses the shortname, shown at all times */
+	/* Map 2 uses the full name, shown at all times */
+	/* Map 3 uses the groupname, shown at all times */
+	if (map == 0 || map == 1) 
+		n = get_server_altname(window->server, 0);
+	else if (map == 3)
+		n = get_server_group(window->server);
 
-	/* Figure out what server this window is on */
-	n = get_server_name(window->server);
-	if (map == 2)
-	{
-		snprintf(my_buffer, 63, server_format, n);
-		return my_buffer;
-	}
+	if (!n)
+		n = get_server_name(window->server);
+	if (!n)
+		return "Unknown";
 
-	name = LOCAL_COPY(n);
+	PRESS(server_format, n)
+	RETURN
+}
 
-	/*
-	 * If the first segment before the first dot is a number,
-	 * then its an ip address, and use the whole thing.
-	 */
-	if (strtoul(name, &next, 10) && *next == '.')
-	{
-		snprintf(my_buffer, 63, server_format, name);
-		return my_buffer;
-	}
+/*
+ * Displays the 005 "NETWORK" value for the current server for the window.
+ */
+STATUS_FUNCTION(status_network)
+{
+	const char *text = NULL;
 
-	/*
-	 * Get the stuff to the left of the first dot.
-	 */
-	if (!(rest = strchr(name, '.')))
-	{
-		snprintf(my_buffer, 63, server_format, name);
-		return my_buffer;
-	}
+	if (window->server != NOSERV)
+		text = get_server_005(window->server, "NETWORK");
 
-	/*
-	 * If the first segment is 'irc', thats not terribly
-	 * helpful, so get the next segment.
-	 */
-	if (!strncmp(name, "irc", 3))
-	{
-		name = rest + 1;
-		if (!(rest = strchr(name + 1, '.')))
-			rest = name + strlen(name);
-	}
-
-	/*
-	 * If the name of the server is > 60 chars, crop it back to 60.
-	 */
-	if ((len = rest - name) > 60)
-		len = 60;
-
-	/*
-	 * Plop the server into the server_format and return it.
-	 */
-	name[len] = 0;
-	snprintf(my_buffer, 63, server_format, name);
-	return my_buffer;
+	return text ? text : "Unknown";
 }
 
 /*
@@ -929,16 +983,14 @@ static	char	my_buffer[64];
  */
 STATUS_FUNCTION(status_query_nick)
 {
-	static char my_buffer[BIG_BUFFER_SIZE + 1];
+	STATUS_VARS
+	const char *q;
 
-	if (window->query_nick && query_format)
-	{
-		snprintf(my_buffer, BIG_BUFFER_SIZE, 
-				query_format, window->query_nick);
-		return my_buffer;
-	}
+	if (!(q = get_equery_by_refnum(window->refnum)))
+		return empty_string;
 
-	return empty_string;
+	PRESS(query_format, q)
+	RETURN
 }
 
 /*
@@ -961,9 +1013,9 @@ STATUS_FUNCTION(status_right_justify)
  */
 STATUS_FUNCTION(status_notify_windows)
 {
+	STATUS_VARS
 	int	doneone = 0;
-	char	buf2[81];
-static	char	my_buffer[81];
+	char	buf2[BIG_BUFFER_SIZE];
 
 	/*
 	 * This only goes to a current-type window.
@@ -971,22 +1023,26 @@ static	char	my_buffer[81];
 	if (!DISPLAY_ON_WINDOW)
 		return empty_string;
 
-	*my_buffer = 0;
-	*buf2 = 0;
-
 	/*
 	 * Look for any notifying windows that have had some output since 
 	 * they have been hidden and collect their refnums.
 	 */
+	*buf2 = 0;
 	window = NULL;
 	while (traverse_all_windows(&window))
 	{
-		if (window->miscflags & WINDOW_NOTIFIED)
+		if (window->notified)
 		{
+			char *s = NULL;
+
+			if (window->notify_name)
+				s = window->notify_name;
+			else if (window->name)
+				s = window->name;
+
 			if (doneone++)
 				strlcat(buf2, ",", sizeof buf2);
-			strlcat(buf2, (map == 1 && window->name) ? 
-					window->name :
+			strlcat(buf2, (map == 1 && s) ? s :
 					ltoa(window->refnum), sizeof buf2);
 		}
 	}
@@ -994,10 +1050,11 @@ static	char	my_buffer[81];
 	/*
 	 * Only do the snprintf if there are windows to process.
 	 */
-	if (doneone && notify_format)
-		snprintf(my_buffer, sizeof my_buffer, notify_format, buf2);
+	if (!doneone)
+		return empty_string;
 
-	return my_buffer;
+	PRESS(notify_format, buf2)
+	RETURN
 }
 
 /*
@@ -1005,14 +1062,13 @@ static	char	my_buffer[81];
  */
 STATUS_FUNCTION(status_clock)
 {
-	static	char	my_buffer[81];
+	STATUS_VARS
 
-	if (get_int_var(CLOCK_VAR) && clock_format && DISPLAY_ON_WINDOW)
-		snprintf(my_buffer, 80, clock_format, get_clock());
-	else
-		*my_buffer = 0;
+	if (!get_int_var(CLOCK_VAR) || !DISPLAY_ON_WINDOW)
+		return empty_string;
 
-	return my_buffer;
+	PRESS(clock_format, get_clock())
+	RETURN
 }
 
 /*
@@ -1020,44 +1076,49 @@ STATUS_FUNCTION(status_clock)
  */
 STATUS_FUNCTION(status_mode)
 {
+	STATUS_VARS
 	const char *	mode = NULL;
 	const char *	chan = NULL;
-static  char    	my_buffer[81];
-
-	/* If anything goes wrong, we return an empty string */
-	*my_buffer = 0;
 
 	/* If the user has no mode format, or we're not connected, punt. */
-        if (!mode_format || window->server == NOSERV)
-		return my_buffer;
+        if (window->server == NOSERV)
+		return empty_string;
 
 	/* If there is a current channel, get it's mode */
 	if ((chan = get_echannel_by_refnum(window->refnum)))
 		mode = get_channel_mode(chan, window->server);
-
-	/* If this is %+, and there is not a mode, punt. */
-	if (mode == 0 && (!mode || !*mode))
-		return my_buffer;
-
-	/* If this is %{1}+, and there is not a mode, use empty string */
 	if (!mode)
 		mode = empty_string;
 
-	/*
-	 * This gross hack is required to make sure that the 
-	 * channel key doesnt accidentally contain anything 
-	 * dangerous...
-	 */
-	if (get_int_var(STATUS_DOES_EXPANDOS_VAR))
+	/* If this is %+, and there is not a mode, punt. */
+	if ((map == 0 || map == 2) && !*mode)
+		return empty_string;
+
+	if (map == 0 || map == 1)
 	{
+	    /*
+	     * This gross hack is required to make sure that the 
+	     * channel key doesnt accidentally contain anything 
+	     * dangerous...
+	     */
+	    if (get_int_var(STATUS_DOES_EXPANDOS_VAR) && strchr(mode, '$'))
+	    {
 		char *mode2 = alloca(strlen(mode) * 2 + 1);
 		double_quote(mode, "$", mode2);
 		mode = mode2;
+	    }
+	}
+	else if (map == 2 || map == 3)
+	{
+		char *x;
+		x = LOCAL_COPY(mode);
+		if (!(mode = next_arg(x, &x)))
+			mode = empty_string;
 	}
 
 	/* Press the mode into the status format. */
-	snprintf(my_buffer, 80, mode_format, mode);
-	return my_buffer;
+	PRESS(mode_format, mode)
+	RETURN
 }
 
 
@@ -1066,8 +1127,8 @@ static  char    	my_buffer[81];
  */
 STATUS_FUNCTION(status_umode)
 {
+	STATUS_VARS
 	char	localbuf[20];
-static	char	my_buffer[81];
 
 	/*
 	 * If we are only on one server and this isnt the current-type 
@@ -1086,8 +1147,8 @@ static	char	my_buffer[81];
 	if (!*localbuf)
 		return empty_string;
 
-	snprintf(my_buffer, 80, umode_format, localbuf);
-	return my_buffer;
+	PRESS(umode_format, localbuf)
+	RETURN
 }
 
 /*
@@ -1136,8 +1197,8 @@ STATUS_FUNCTION(status_ssl)
  */
 STATUS_FUNCTION(status_hold_lines)
 {
+	STATUS_VARS
 	int	num;
-static	char	my_buffer[81];
 	int	interval = window->hold_interval;
 	int	lines_held;
 
@@ -1145,20 +1206,18 @@ static	char	my_buffer[81];
 		interval = 1;		/* XXX WHAT-ever */
 
 	if (window->holding_distance_from_display_ip > window->scrollback_distance_from_display_ip)
-		lines_held = window->holding_distance_from_display_ip - window->display_size;
+		lines_held = window->holding_distance_from_display_ip - window->display_lines;
 	else
-		lines_held = window->scrollback_distance_from_display_ip - window->display_size;
+		lines_held = window->scrollback_distance_from_display_ip - window->display_lines;
 
 	if (lines_held <= 0)
 		return empty_string;
 
-	if ((num = (lines_held / interval) * interval))
-	{
-		snprintf(my_buffer, 80, hold_lines_format, ltoa(num));
-		return my_buffer;
-	}
+	if ((num = (lines_held / interval) * interval) == 0)
+		return empty_string;
 
-	return empty_string;
+	PRESS(hold_lines_format, ltoa(num))
+	RETURN
 }
 
 /*
@@ -1166,9 +1225,9 @@ static	char	my_buffer[81];
  */
 STATUS_FUNCTION(status_channel)
 {
+	STATUS_VARS
 	const char *chan;
 	char 	channel[IRCD_BUFFER_SIZE + 1];
-static	char	my_buffer[IRCD_BUFFER_SIZE + 1];
 	int	num;
 
 	if (window->server == NOSERV || !channel_format)
@@ -1187,9 +1246,8 @@ static	char	my_buffer[IRCD_BUFFER_SIZE + 1];
 	if (num > 0 && (int)strlen(channel) > num)
 		channel[num] = 0;
 
-	snprintf(my_buffer, IRCD_BUFFER_SIZE, 
-			channel_format, check_channel_type(channel));
-	return my_buffer;
+	PRESS(channel_format, check_channel_type(channel))
+	RETURN
 }
 
 /*
@@ -1219,8 +1277,8 @@ STATUS_FUNCTION(status_voice)
  */
 STATUS_FUNCTION(status_mail)
 {
+	STATUS_VARS
 	const char *	number;
-static	char	my_buffer[81];
 
 	/*
 	 * The order is important here.  We check to see whether or not
@@ -1228,14 +1286,12 @@ static	char	my_buffer[81];
 	 * that will be false, and check_mail() is very expensive; we dont
 	 * want to do it if we're going to ignore the result.
 	 */
-	if (get_int_var(MAIL_VAR) && mail_format && 
-	    DISPLAY_ON_WINDOW && (number = check_mail()))
-	{
-		snprintf(my_buffer, 80, mail_format, number);
-		return my_buffer;
-	}
+	if (!get_int_var(MAIL_VAR) || !DISPLAY_ON_WINDOW || 
+				((number = check_mail()) == 0))
+		return empty_string;
 
-	return empty_string;
+	PRESS(mail_format, number)
+	RETURN
 }
 
 /*
@@ -1300,32 +1356,32 @@ STATUS_FUNCTION(status_user)
 
 	/* XXX Ick.  Oh well. */
 	struct dummystruct {
-		int		map;
+		short		map;
 		char		key;
-		enum VAR_TYPES	var;
+		int *		var;
 	} lookup[] = {
-	{ 0, 'U', STATUS_USER0_VAR }, { 0, 'X', STATUS_USER1_VAR },
-	{ 0, 'Y', STATUS_USER2_VAR }, { 0, 'Z', STATUS_USER3_VAR },
-	{ 0, '0', STATUS_USER0_VAR }, { 0, '1', STATUS_USER1_VAR },
-	{ 0, '2', STATUS_USER2_VAR }, { 0, '3', STATUS_USER3_VAR },
-	{ 0, '4', STATUS_USER4_VAR }, { 0, '5', STATUS_USER5_VAR },
-	{ 0, '6', STATUS_USER6_VAR }, { 0, '7', STATUS_USER7_VAR },
-	{ 0, '8', STATUS_USER8_VAR }, { 0, '9', STATUS_USER9_VAR },
-	{ 1, '0', STATUS_USER10_VAR }, { 1, '1', STATUS_USER11_VAR },
-	{ 1, '2', STATUS_USER12_VAR }, { 1, '3', STATUS_USER13_VAR },
-	{ 1, '4', STATUS_USER14_VAR }, { 1, '5', STATUS_USER15_VAR },
-	{ 1, '6', STATUS_USER16_VAR }, { 1, '7', STATUS_USER17_VAR },
-	{ 1, '8', STATUS_USER18_VAR }, { 1, '9', STATUS_USER19_VAR },
-	{ 2, '0', STATUS_USER20_VAR }, { 2, '1', STATUS_USER21_VAR },
-	{ 2, '2', STATUS_USER22_VAR }, { 2, '3', STATUS_USER23_VAR },
-	{ 2, '4', STATUS_USER24_VAR }, { 2, '5', STATUS_USER25_VAR },
-	{ 2, '6', STATUS_USER26_VAR }, { 2, '7', STATUS_USER27_VAR },
-	{ 2, '8', STATUS_USER28_VAR }, { 2, '9', STATUS_USER29_VAR },
-	{ 3, '0', STATUS_USER30_VAR }, { 3, '1', STATUS_USER31_VAR },
-	{ 3, '2', STATUS_USER32_VAR }, { 3, '3', STATUS_USER33_VAR },
-	{ 3, '4', STATUS_USER34_VAR }, { 3, '5', STATUS_USER35_VAR },
-	{ 3, '6', STATUS_USER36_VAR }, { 3, '7', STATUS_USER37_VAR },
-	{ 3, '8', STATUS_USER38_VAR }, { 3, '9', STATUS_USER39_VAR },
+	{ 0, 'U', &STATUS_USER_VAR }, { 0, 'X', &STATUS_USER1_VAR },
+	{ 0, 'Y', &STATUS_USER2_VAR }, { 0, 'Z', &STATUS_USER3_VAR },
+	{ 0, '0', &STATUS_USER_VAR }, { 0, '1', &STATUS_USER1_VAR },
+	{ 0, '2', &STATUS_USER2_VAR }, { 0, '3', &STATUS_USER3_VAR },
+	{ 0, '4', &STATUS_USER4_VAR }, { 0, '5', &STATUS_USER5_VAR },
+	{ 0, '6', &STATUS_USER6_VAR }, { 0, '7', &STATUS_USER7_VAR },
+	{ 0, '8', &STATUS_USER8_VAR }, { 0, '9', &STATUS_USER9_VAR },
+	{ 1, '0', &STATUS_USER10_VAR }, { 1, '1', &STATUS_USER11_VAR },
+	{ 1, '2', &STATUS_USER12_VAR }, { 1, '3', &STATUS_USER13_VAR },
+	{ 1, '4', &STATUS_USER14_VAR }, { 1, '5', &STATUS_USER15_VAR },
+	{ 1, '6', &STATUS_USER16_VAR }, { 1, '7', &STATUS_USER17_VAR },
+	{ 1, '8', &STATUS_USER18_VAR }, { 1, '9', &STATUS_USER19_VAR },
+	{ 2, '0', &STATUS_USER20_VAR }, { 2, '1', &STATUS_USER21_VAR },
+	{ 2, '2', &STATUS_USER22_VAR }, { 2, '3', &STATUS_USER23_VAR },
+	{ 2, '4', &STATUS_USER24_VAR }, { 2, '5', &STATUS_USER25_VAR },
+	{ 2, '6', &STATUS_USER26_VAR }, { 2, '7', &STATUS_USER27_VAR },
+	{ 2, '8', &STATUS_USER28_VAR }, { 2, '9', &STATUS_USER29_VAR },
+	{ 3, '0', &STATUS_USER30_VAR }, { 3, '1', &STATUS_USER31_VAR },
+	{ 3, '2', &STATUS_USER32_VAR }, { 3, '3', &STATUS_USER33_VAR },
+	{ 3, '4', &STATUS_USER34_VAR }, { 3, '5', &STATUS_USER35_VAR },
+	{ 3, '6', &STATUS_USER36_VAR }, { 3, '7', &STATUS_USER37_VAR },
+	{ 3, '8', &STATUS_USER38_VAR }, { 3, '9', &STATUS_USER39_VAR },
 	{ 0, 0, 0 },
 	};
 
@@ -1334,7 +1390,7 @@ STATUS_FUNCTION(status_user)
 	{
 		if (map == lookup[i].map && key == lookup[i].key)
 		{
-			text = get_string_var(lookup[i].var);
+			text = get_string_var(*lookup[i].var);
 			break;
 		}
 	}
@@ -1349,13 +1405,33 @@ STATUS_FUNCTION(status_hold)
 {
 	char *text;
 
-	if (window->holding_distance_from_display_ip > window->display_size  ||
-	    window->scrollback_distance_from_display_ip > window->display_size)
+	if (window->holding_distance_from_display_ip > window->display_lines  ||
+	    window->scrollback_distance_from_display_ip > window->display_lines)
 		if ((text = get_string_var(STATUS_HOLD_VAR)))
 			return text;
 	return empty_string;
 }
 
+STATUS_FUNCTION(status_holdmode)
+{
+	char *text;
+
+	/* If hold mode is on... */
+	if (window->holding_top_of_display)
+	{
+	    /* ... and we're not holding anything */
+	    if (window->holding_distance_from_display_ip <
+				window->display_lines  &&
+	         window->scrollback_distance_from_display_ip <
+				window->display_lines)
+	    {
+		if ((text = get_string_var(STATUS_HOLDMODE_VAR)))
+			return text;
+	    }
+	}
+	return empty_string;
+}
+
 STATUS_FUNCTION(status_oper)
 {
 	char *text;
@@ -1370,30 +1446,48 @@ STATUS_FUNCTION(status_oper)
 
 STATUS_FUNCTION(status_window)
 {
-	char *text;
+	const char *text;
 
-	if ((number_of_windows_on_screen(window) > 1) && 
-	    IS_CURRENT_WINDOW && (text = get_string_var(STATUS_WINDOW_VAR)))
-		return text;
-	else
-		return empty_string;
+	if (!(text = get_string_var(STATUS_WINDOW_VAR)))
+		text = empty_string;
+
+	switch (map)
+	{
+		case 0:
+			if (number_of_windows_on_screen(window) <= 1)
+				break;
+			/* FALLTHROUGH */
+		case 3:
+			if (!IS_CURRENT_WINDOW)
+				break;
+			/* FALLTHROUGH */
+		case 2:
+			return text;
+	}
+
+	return empty_string;
 }
 
 STATUS_FUNCTION(status_refnum)
 {
-	static char my_buffer[81];
+	STATUS_VARS
+	const char *value;
 
-	strlcpy(my_buffer, window->name ? window->name 
-					: ltoa(window->refnum), sizeof my_buffer);
-	return my_buffer;
+	if (window->name)
+		value = window->name;
+	else
+		value = ltoa(window->refnum);
+
+	PRESS("%s", value)
+	RETURN
 }
 
 STATUS_FUNCTION(status_refnum_real)
 {
-	static char my_buffer[81];
+	STATUS_VARS
 
-	strlcpy(my_buffer, ltoa(window->refnum), sizeof my_buffer);
-	return my_buffer;
+	PRESS("%s", ltoa(window->refnum))
+	RETURN
 }
 
 STATUS_FUNCTION(status_version)
@@ -1417,8 +1511,9 @@ STATUS_FUNCTION(status_null_function)
  */
 STATUS_FUNCTION(status_dcc)
 {
-	if ((current_window_level & LOG_DCC && IS_CURRENT_WINDOW) ||
-			(window->window_level & LOG_DCC))
+	if ((mask_isset(&current_window_mask, LEVEL_DCC) && 
+				IS_CURRENT_WINDOW) ||
+	    (mask_isset(&window->window_mask, LEVEL_DCC)))
 		return DCC_get_current_transfer();
 
 	return empty_string;
@@ -1439,15 +1534,13 @@ STATUS_FUNCTION(status_dcc_all)
  */
 STATUS_FUNCTION(status_cpu_saver_mode)
 {
-	static char my_buffer[81];
+	STATUS_VARS
 
-	if (cpu_saver && cpu_saver_format)
-	{
-		snprintf(my_buffer, 80, cpu_saver_format, "CPU");
-		return my_buffer;
-	}
+	if (!cpu_saver)
+		return empty_string;
 
-	return empty_string;
+	PRESS(cpu_saver_format, "CPU")
+	RETURN
 }
 
 /*
@@ -1459,15 +1552,11 @@ STATUS_FUNCTION(status_position)
 {
 	static char my_buffer[81];
 
-	snprintf(my_buffer, 80, "(%d/%d/%d-%d-%d)", 
+	snprintf(my_buffer, sizeof my_buffer, "(%d/%d/%d-%d-%d)", 
 			window->scrolling_distance_from_display_ip,
 			window->holding_distance_from_display_ip,
 			window->scrollback_distance_from_display_ip,
-			window->display_size, window->cursor);
-#if 0
-			window->screen->input_line,
-			window->screen->input_cursor);
-#endif
+			window->display_lines, window->cursor);
 	return my_buffer;
 }
 
@@ -1490,29 +1579,22 @@ STATUS_FUNCTION(status_scroll_info)
 {
 	static char my_buffer[81];
 
-	if (window->scrollback_top_of_display)
-	{
-		snprintf(my_buffer, 80, " (Scroll: %d of %d)", 
-				window->scrollback_distance_from_display_ip,
-				window->display_buffer_size - 1);
-	}
-	else
-		*my_buffer = 0;
+	if (!window->scrollback_top_of_display)
+		return empty_string;
 
+	snprintf(my_buffer, sizeof my_buffer, " (Scroll: %d of %d)", 
+			window->scrollback_distance_from_display_ip,
+			window->display_buffer_size - 1);
 	return my_buffer;
 }
 
 
 STATUS_FUNCTION(status_windowspec)
 {
-	static char my_buffer[81];
-
 	if (window->status.special)
-		strlcpy(my_buffer, window->status.special, sizeof my_buffer);
+		return window->status.special;
 	else
-		*my_buffer = 0;
-
-	return my_buffer;
+		return empty_string;
 }
 
 STATUS_FUNCTION(status_percent)
@@ -1542,4 +1624,30 @@ STATUS_FUNCTION(status_swappable)
 		return empty_string;
 }
 
+STATUS_FUNCTION(status_activity)
+{
+	const char *format, *data;
+	char *result;
+	static char retval[80];
+
+	if (window->current_activity == 0)
+		return empty_string;
+
+	format = window->activity_format[window->current_activity];
+	if (!format || !*format)
+		format = window->activity_format[0];
+	if (!format || !*format)
+		return empty_string;
+
+	data = window->activity_data[window->current_activity];
+	if (!data || !*data)
+		data = window->activity_data[0];
+	if (!data || !*data)
+		data = empty_string;
+
+	result = expand_alias(format, data);
+	strlcpy(retval, result, sizeof(retval));
+	new_free(&result);
+	return retval;
+}
 
